#include "rannuMain.h"

extern char *random_number;
extern long untouched;
extern MANAGE_MEMO *mm;
extern ManagePane *managePane;
extern Rannu *rannu;

namespace {
    using namespace std;          

    const string m0(          ""); // m[0], FULL_PATH_NAME includes MANAGE_TEXT.
    const string m1(         "0"); // m[1], csv: RANNU_CSV
    const string m2(      "C:\\"); // m[2], pathName includes the directory name (C:\) and RANNU_TEXT.
    const string m3(        "32"); // m[3], kilo: RANNU_TEXT size (kilobytes, KB)
    const string m4(         "1"); // m[4], interval_0: Loto-7
    const string m5(        "37"); // m[5], interval_1: Loto-7
    const string m6(         "0"); // m[6], grouping: interval of 0-63
    const string m7(       "400"); // m[7], from: idling
    const string m8(    "123456"); // m[8], seed: pseudorandom number generator (PRNG)
    const string m9("2147483647"); // m[9]
}

bool ManagePane::readManageText() {
    using namespace std;

    char c[BUFFER_SIZE];
    string *p = new string("");
    ifstream in( FULL_PATH_NAME.c_str() );
    if( in.fail() ) {
        error = "e14: ifstream in( " + FULL_PATH_NAME + CLOSE_BRACKET;
        createErrorText(error);
        return false;
    }

    m[0] = FULL_PATH_NAME;
    for(int line = 1; line <= MANAGE_TEXT_LINE_COUNT; line += 1) {
        if( in.getline(c, BUFFER_SIZE) ) {
            delete p;
            p = new string(c);
            m[line] = *p;
        } else {
            break;
        }

    }

    in.close();
    delete p;
    return true;
}

bool ManagePane::writeManageText() {
    using namespace std;

    char *stop;
    const int DECIMAL = 10;
    int f;
    long k;

    m[0] = FULL_PATH_NAME;
    m[1] = csv ? ONE : ZERO;

    trimming( text_buffer->text(), buffer);
    string PATH_NAME(buffer);
    if(csv) {
        pathName = PATH_NAME + RANNU_CSV;
    } else {
        pathName = PATH_NAME + RANNU_TEXT;
    }

    m[2] = PATH_NAME; // root + directory

    trimming( input2->value(), buffer);
    string KILO(buffer);
    k = strtol(buffer, &stop, DECIMAL);
    kilo = k >= 0 ? k : 0L;
    m[3] = KILO;

    trimming( input4->value(), buffer);
    string INTERVAL_0(buffer);
    f = (int)strtol(buffer, &stop, DECIMAL);
    interval_0 = f >= 0 ? f : 0;
    m[4] = INTERVAL_0;

    trimming( input5->value(), buffer);
    string INTERVAL_1(buffer);
    f = (int)strtol(buffer, &stop, DECIMAL);
    interval_1 = f >= 0 ? f : 0;
    m[5] = INTERVAL_1;

    grouping = false;
    if(interval_1 == 63) {
        if(interval_0 == 0) {
            grouping = true;
        }

    }

    mm->grouping = grouping;
    m[6] = grouping ? ONE : ZERO;

    trimming( input6->value(), buffer);
    string FROM(buffer);
    f = (int)strtol(buffer, &stop, DECIMAL);
    from = f >= 0 ? f : 0; // idling
    m[7] = FROM;

    trimming( input7->value(), buffer);
    string SEED(buffer);
    k = strtol(buffer, &stop, DECIMAL);
    seed = k >= 0 ? k : 0L;
    m[8] = SEED;

    trimming( input3->value(), buffer);
    string LINE_COUNT(buffer);
    f = (int)strtol(buffer, &stop, DECIMAL);
    lineCount = f >= 0 ? f : 0;
    m[9] = LINE_COUNT;

// Creating MANAGE_TEXT
    ofstream manage_text( FULL_PATH_NAME.c_str() );
    for(int line = 1; line <= MANAGE_TEXT_LINE_COUNT; line += 1 ) {
        manage_text << m[line] << endl;
    }

    return true;
}

char* ManagePane::right_justifying(char *p) {
    char main_buffer[MAIN_BUFFER_SIZE];
    const int TEXT_FIELD_SIZE = 32;
    const int LARGE_TEXT_FIELD_SIZE = 77;
    int p_length = strlen(p);
    int textFieldSize;

    if(p_length > LARGE_TEXT_FIELD_SIZE) {
        display->textsize(8);
        return p;
    } else if(p_length > TEXT_FIELD_SIZE) {
        display->textsize(9);
        textFieldSize = LARGE_TEXT_FIELD_SIZE;
    } else {
        display->textsize(20);
        textFieldSize = TEXT_FIELD_SIZE;
    }

    strcpy(main_buffer, "                                                                                "); // space width of 80
    strcat(main_buffer, p);

    const int SPACE_DIRECTORY_SIZE = strlen(main_buffer);
    const int MOVE = SPACE_DIRECTORY_SIZE - textFieldSize;
    char *moved = &main_buffer[MOVE];

    return moved;
}

char* ManagePane::right_justifying(int value) {
    using namespace std;

    int s_length;
    string s;
    stringstream os;
    os << value;
    s = os.str();
    s_length = s.length();

    switch(s_length) {
    case 1:
        strcpy(buffer, "                  ");
        strcat(buffer, s.c_str() );
        break;
    case 2:
        strcpy(buffer, "                ");
        strcat(buffer, s.c_str() );
        break;
    case 3:
        strcpy(buffer, "              ");
        strcat(buffer, s.c_str() );
        break;
    case 4:
        strcpy(buffer, "            ");
        strcat(buffer, s.c_str() );
        break;
    case 5:
        strcpy(buffer, "          ");
        strcat(buffer, s.c_str() );
        break;
    case 6:
        strcpy(buffer, "         ");
        strcat(buffer, s.c_str() );
        break;
    case 7:
        strcpy(buffer, "        ");
        strcat(buffer, s.c_str() );
        break;
    case 8:
        strcpy(buffer, "      ");
        strcat(buffer, s.c_str() );
        break;
    case 9:
        strcpy(buffer, "    ");
        strcat(buffer, s.c_str() );
        break;
    case 10:
        strcpy(buffer, "  ");
        strcat(buffer, s.c_str() );
        break;
    case 11:
        strcpy(buffer, s.c_str() );
        break;
    }

    return &buffer[0];
}

char* ManagePane::right_justifying(long v) {
    int value = (int)v;
    right_justifying(value);
}

ManagePane::ManagePane(int W, int H, const char *t) : Fl_Double_Window(W, H, t), FULL_PATH_NAME("C:\\rannuFLTK\\manage.txt"), CLOSE_BRACKET(" )"), NEW_LINE("\x0A"), ONE("1"), RANNU_CSV("rannu.csv"), RANNU_TEXT("rannu.txt"), ZERO("0") {
    m[0] = m0; m[1] = m1; m[2] = m2; m[3] = m3; m[4] = m4; m[5] = m5; m[6] = m6; m[7] = m7; m[8] = m8; m[9] = m9; 
}

NAMES driveDirectoryBase(std::string & s) {
    using namespace std;

    const string EMPTY = "";
    const string SLASH = "/";
    const string YEN = "\\";
    int firstTrace, trace, w;
    int s_length = s.length();
    string c;

    NAMES part, undivided;
    part.name[0] = EMPTY; part.name[1] = EMPTY; part.name[2] = EMPTY; part.name[3] = EMPTY;
    undivided.name[0] = EMPTY; undivided.name[1] = EMPTY; undivided.name[2] = EMPTY; undivided.name[3] = EMPTY;

    part.name[0] = s; // s.length > 0; in other words, part.name[0].length > undivided.name[0].length

// s is divided into three parts: drive name, directory name, and base name.
    int k = 0;
    for( ; k < s_length; ++k) {
        c = s.substr(k, 1);
        if(c == YEN) {
            part.separator = YEN;
            break;
        } else if(c == SLASH) {
            part.separator = SLASH;
            break;
        }

    }

    if(k == s_length) { // Neither YEN nor SLASH has been found.
        return undivided;
    } else {
        w = k + 1;
        part.name[1] = s.substr(0, w); // C:\ for example
        trace = ++k;
        firstTrace = trace;
    }

    for( ; k < s_length; ++k) {
        c = s.substr(k, 1);
        if(c == part.separator) {
            trace = ++k;
        }

    }

    w = trace - firstTrace;
    part.name[2] = s.substr(firstTrace, w);
    part.name[3] = s.substr(trace); // manage.txt for example
    return part;
}

void close_manage_cb(Fl_Widget *, void *) {
    managePane->hide();
}

void figures(char *buffer, int value) {
    using namespace std;

    stringstream os;
    os << value;
    string s = os.str();
    strcpy(buffer, s.c_str() );
}

void figures(char *buffer, long value) {
    using namespace std;

    stringstream os;
    os << value;
    string s = os.str();
    strcpy(buffer, s.c_str() );
}

void focus_cb(Fl_Widget *, void *) {
    managePane->focusHandler();
}

void KB_cb(Fl_Widget *, void *) {
    managePane->KBhandler();
}

void line_count_cb(Fl_Widget *, void *) {
    managePane->lineCountHandler();
}

void ManagePane::fileOpenDialog() {
    using namespace std;

    Fl_File_Chooser chooser(".", Fl_File_Chooser::all_files_label, Fl_File_Chooser::SINGLE, "Where");
    chooser.show();

    while( chooser.shown() ) {
        Fl::wait();
    }

    if( chooser.value() == NULL ) {
        return; 
    }

    string pathName( chooser.value() );

    NAMES a = driveDirectoryBase(pathName);
    string root = a.name[1];
    string directory = a.name[2];
    path = root + directory;

    strcpy(buffer, path.c_str() );
    replaceSlashWithYen(buffer);
    text_buffer->text( right_justifying(buffer) );
}

void ManagePane::focusHandler() {
    saveButton->take_focus();
}

void ManagePane::KBhandler() {
    char *stop;
    const int DECIMAL = 10;

    trimming( input2->value(), buffer);
    string KILO(buffer);
    long k = strtol(buffer, &stop, DECIMAL);
    kilo = k >= 0 ? k : 0L;
    m[3] = KILO;

    mm->kilo = kilo;
    input2->value( right_justifying(kilo) );
}

void ManagePane::lineCountHandler() {
    char *stop;
    const int DECIMAL = 10;

    trimming( input3->value(), buffer);
    string LINE_COUNT(buffer);
    int f = (int)strtol(buffer, &stop, DECIMAL);
    lineCount = f >= 0 ? f : 0;
    m[9] = LINE_COUNT;

    mm->lineCount = lineCount;
    input3->value( right_justifying(lineCount) );
}

void ManagePane::rannuCsvHandler() {
    rannuCsvButton->hide();
    rannuTextButton->hide();

    csv = false;
    rannuTextButton->show();

    mm->csv = csv;
}

void ManagePane::rannuTextHandler() {
    rannuCsvButton->hide();
    rannuTextButton->hide();

    csv = true;
    rannuCsvButton->show();

    mm->csv = csv;
}

void ManagePane::saveHandler() {
    if( readManageText() ) {
        writeManageText();
    }

}

void ManagePane::setting(bool building) {
    using namespace std;

    if(building) {
        const int ADJUSTER = 4;
        const int NARROW = 1;
        const int MANAGE_RANNU_CSV_BUTTON_WIDTH = MANAGE_RANNU_TEXT_BUTTON_WIDTH;
        int x, y;

        whereButton = new Fl_Button(MANAGE_WHERE_BUTTON_LEFT, MANAGE_WHERE_BUTTON_TOP, BUTTON_WIDTH, BUTTON_HEIGHT, "Where");
        whereButton->callback( (Fl_Callback *)where_cb, this);

        saveButton = new Fl_Button(MANAGE_BUTTON_LEFT, MANAGE_SAVE_BUTTON_TOP, BUTTON_WIDTH, BUTTON_HEIGHT, "Save");
        saveButton->callback( (Fl_Callback *)save_cb, this);

        closeButton = new Fl_Button(MANAGE_BUTTON_LEFT, MANAGE_CLOSE_BUTTON_TOP, BUTTON_WIDTH, BUTTON_HEIGHT, "Close");
        closeButton->callback( (Fl_Callback *)close_manage_cb, this);

        helpButton = new Fl_Button(MANAGE_BUTTON_LEFT, MANAGE_HELP_BUTTON_TOP, BUTTON_WIDTH, BUTTON_HEIGHT, "Help");
        helpButton->callback( (Fl_Callback *)help_cb, this);

        x = MANAGE_DIRECTORY_LEFT + MANAGE_DIRECTORY_WIDTH + NARROW + ADJUSTER;
        y = MANAGE_DIRECTORY_TOP;
        rannuCsvButton = new Fl_Button(x, y, MANAGE_RANNU_CSV_BUTTON_WIDTH, MANAGE_DIRECTORY_HEIGHT, "rannu.csv");
        rannuCsvButton->callback( (Fl_Callback *)rannu_csv_cb, this);
        rannuCsvButton->box(FL_FLAT_BOX);

        x = MANAGE_DIRECTORY_LEFT + MANAGE_DIRECTORY_WIDTH + NARROW;
        y = MANAGE_DIRECTORY_TOP;
        rannuTextButton = new Fl_Button(x, y, MANAGE_RANNU_TEXT_BUTTON_WIDTH, MANAGE_DIRECTORY_HEIGHT, "rannu.txt");
        rannuTextButton->callback( (Fl_Callback *)rannu_text_cb, this);
        rannuTextButton->box(FL_FLAT_BOX);

        x = MANAGE_KB_LEFT + MANAGE_KB_WIDTH + NARROW;
        y = MANAGE_KB_TOP;
        kilobytesButton = new Fl_Button(x, y, MANAGE_KILOBYTES_BUTTON_WIDTH, MANAGE_DIRECTORY_HEIGHT, "KB");
        kilobytesButton->callback( (Fl_Callback *)KB_cb, this);
        kilobytesButton->box(FL_FLAT_BOX);

        x = MANAGE_LINES_LEFT + MANAGE_LINES_WIDTH + NARROW;
        y = MANAGE_LINES_TOP;
        linesButton = new Fl_Button(x, y, MANAGE_LINES_BUTTON_WIDTH, MANAGE_DIRECTORY_HEIGHT, "lines");
        linesButton->callback( (Fl_Callback *)line_count_cb, this);
        linesButton->box(FL_FLAT_BOX);

        prngButton = new Fl_Button(MANAGE_PRNG_LEFT, MANAGE_PRNG_TOP, MANAGE_PRNG_WIDTH, MANAGE_PRNG_HEIGHT, "");
        prngButton->callback( (Fl_Callback *)focus_cb, this);
        prngButton->box(FL_ENGRAVED_FRAME); // group box of PRNG

        x = MANAGE_PRNG_LEFT + 11;
        y = MANAGE_PRNG_TOP - 13;
        captionButton = new Fl_Button(x, y, MANAGE_CAPTION_BUTTON_WIDTH, MANAGE_CAPTION_BUTTON_HEIGHT, "PRNG");
        captionButton->box(FL_FLAT_BOX); // caption of the group box of PRNG

        text_buffer = new Fl_Text_Buffer();
        display = new Fl_Text_Display(MANAGE_DIRECTORY_LEFT, MANAGE_DIRECTORY_TOP, MANAGE_DIRECTORY_WIDTH, MANAGE_DIRECTORY_HEIGHT, "");
        display->buffer(text_buffer);
        display->textfont(FL_COURIER);
        display->textsize(20);
        display->box(FL_BORDER_BOX);

        input2 = new Fl_Input(MANAGE_KB_LEFT, MANAGE_KB_TOP, MANAGE_KB_WIDTH, MANAGE_KB_HEIGHT, "");
        input2->box(FL_BORDER_BOX);
        input2->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );

        input3 = new Fl_Input(MANAGE_LINES_LEFT, MANAGE_LINES_TOP, MANAGE_LINES_WIDTH, MANAGE_LINES_HEIGHT, "");
        input3->box(FL_BORDER_BOX);
        input3->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );

        x = MANAGE_PRNG_LEFT + MANAGE_FROM_LEFT;
        y = MANAGE_PRNG_TOP + MANAGE_FROM_TOP;
        input4 = new Fl_Input(x, y, MANAGE_FROM_WIDTH, MANAGE_FROM_HEIGHT, "from");
        input4->box(FL_BORDER_BOX);
        input4->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );

        x = MANAGE_PRNG_LEFT + MANAGE_TO_LEFT;
        y = MANAGE_PRNG_TOP + MANAGE_TO_TOP;
        input5 = new Fl_Input(x, y, MANAGE_TO_WIDTH, MANAGE_TO_HEIGHT, "to");
        input5->box(FL_BORDER_BOX);
        input5->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );

        x = MANAGE_PRNG_LEFT + MANAGE_IDLING_LEFT;
        y = MANAGE_PRNG_TOP + MANAGE_IDLING_TOP;
        input6 = new Fl_Input(x, y, MANAGE_IDLING_WIDTH, MANAGE_IDLING_HEIGHT, "idling");
        input6->box(FL_BORDER_BOX);
        input6->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );

        x = MANAGE_PRNG_LEFT + MANAGE_SEED_LEFT;
        y = MANAGE_PRNG_TOP + MANAGE_SEED_TOP;
        input7 = new Fl_Input(x, y, MANAGE_SEED_WIDTH, FIELD_HEIGHT, "seed");
        input7->box(FL_BORDER_BOX);
        input7->when( FL_WHEN_ENTER_KEY | FL_WHEN_NOT_CHANGED );
    }

    saveButton->take_focus();
    csv = mm->csv;
    if(csv) {
        rannuCsvButton->show();
        rannuTextButton->hide();
    } else {
        rannuCsvButton->hide();
        rannuTextButton->show();
    }

    pathName = mm->pathName;
    NAMES a = driveDirectoryBase(pathName);
    string root = a.name[1];
    string directory = a.name[2];
    path = root + directory;

    strcpy(buffer, path.c_str() );
    replaceSlashWithYen(buffer);
    text_buffer->text( right_justifying(buffer) );

    kilo = mm->kilo;
    input2->value( right_justifying(kilo) );

    lineCount = mm->lineCount;
    input3->value( right_justifying(lineCount) );

    interval_0 = mm->interval_0;
    figures(buffer, interval_0);
    input4->value(buffer);

    interval_1 = mm->interval_1;
    figures(buffer, interval_1);
    input5->value(buffer);

    from = mm->from; // idling
    figures(buffer, from);
    input6->value(buffer);

    seed = mm->seed;
    figures(buffer, seed);
    input7->value(buffer);
    return;
}

void rannu_csv_cb(Fl_Widget *, void *) {
    managePane->rannuCsvHandler();
}

void rannu_text_cb(Fl_Widget *, void *) {
    managePane->rannuTextHandler();
}

void replaceSlashWithYen(char *buffer) {
    char *p = buffer;
    char c;
    const char CARRIAGE_RETURN = 0x0D;
    const char LINE_FEED = 0x0A;
    const char SLASH = 0x2F;
    const char YEN = 0x5C;

    for(;;) {
        c = *p;
        if(c == 0) {
            break;
        } else if(c == SLASH) {
            *p = YEN;
        } else if(c == LINE_FEED) {
            break;
        } else if(c == CARRIAGE_RETURN) {
            break;
        }

        p++;
    }

}

void save_cb(Fl_Widget *, void *) {
    managePane->saveHandler();
}

void trimming(const char *t, char *b) {
    int n, r, size;
    int t_length = strlen(t);
    char *p = new char[ t_length + 1 ];
    strcpy(p, t);
    r = strspn( strrev(p), " ");
    n = strspn(t, " ");

    size = t_length - n - r;
    memcpy(b, &t[n], size);
    b[size] = '\0';
    delete p;
}

void where_cb(Fl_Widget *, void *) {
    managePane->fileOpenDialog();
}
